// HTML Cleaner DOMparser, Change Stylesheets to HTML 4 attributs, HTML formatter
// copyright 2023 mirko boeer

var DontCorrectHTMLElementForEMails = false;

        // from https://stackoverflow.com/questions/23223718/failed-to-execute-btoa-on-window-the-string-to-be-encoded-contains-characte
        // none ascii chars can't be base64 encoded
        function btoa_utf8(value) {
            return btoa(
                String.fromCharCode(
                    ...new TextEncoder('utf-8')
                           .encode(value)
                )
            );
        }

        function atob_utf8(value) {
            const value_latin1 = atob(value);
            return new TextDecoder('utf-8').decode(
                Uint8Array.from(
                    { length: value_latin1.length },
                    (element, index) => value_latin1.charCodeAt(index)
                )
            )
        }

function GetEMailCleanedHTML(htmlString, callbackAfterDOMParser, basePath, relativeUserFilesPath){

         if(typeof basePath === "undefined")
           basePath = "";

         if(typeof relativeUserFilesPath === "undefined")
           relativeUserFilesPath = "";

         const Rules=[];

         var tags = ['div', 'p', 'table', 'td', 'tr', 'ol', 'ul', 'li', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'];
         for (var key in tags)
            Rules[tags[key]] = {
                        indent : false,
                        breakBeforeOpen : true,
                        breakAfterOpen : false,
                        breakBeforeClose : false,
                        breakAfterClose : true
                };

         Rules['head'] =
                     {
                         indent : false,
                         breakBeforeOpen : true,
                         breakAfterOpen : true,
                         breakBeforeClose : false,
                         breakAfterClose : false
                     };

         Rules['title'] =
                     {
                         indent : false,
                         breakBeforeOpen : true,
                         breakAfterOpen : false,
                         breakBeforeClose : false,
                         breakAfterClose : true
                     };

         Rules['body'] =
                     {
                         indent : false,
                         breakBeforeOpen : true,
                         breakAfterOpen : true,
                         breakBeforeClose : true,
                         breakAfterClose : false
                     };

         Rules['meta'] =
                     {
                         indent : false,
                         breakBeforeOpen : true,
                         breakAfterOpen : true,
                         breakBeforeClose : true,
                         breakAfterClose : true
                     };
         Rules['style'] =
                     {
                         indent : false,
                         breakBeforeOpen : true,
                         breakAfterOpen : true,
                         breakBeforeClose : false,
                         breakAfterClose : true
                     };

         Rules['br'] =
                     {
                         indent : false,
                         breakBeforeOpen : false,
                         breakAfterOpen : false,
                         breakBeforeClose : false,
                         breakAfterClose : false
                     };


         options = { useRules: true, indentationChars: '    ', Rules: Rules, basePath: basePath, relativeUserFilesPath: relativeUserFilesPath };


        //var htmlString = document.getElementById("htmlContent").innerHTML;
        //htmlString = htmlString.replace(/&lt;/g, '<').replace(/&gt;/g, '>'); // replace < > from entities;

        // insert <title>
        var pos = htmlString.indexOf('<head>');
        if(  pos > 0 && htmlString.indexOf('<title>') < 0 ){
           let ins_string = "<title></title>";
           htmlString = htmlString.slice(0, pos + 6) + ins_string + htmlString.slice(pos + 6);
        }

        // remove caret marker when it is in comments
        htmlString = htmlString.replace(/\<span id\=\"CmCaReT\"\>\<\/span\>/ig, "");

        // remove whitespaces and protect comments/pre/&nbsp; => simpler DOMParser usage without textnodes
        htmlString = htmlString.replace(/((?:<\!--))([\s\S]*?)((?:-->))/g, function(match, p1, p2, p3) { return p1 + btoa_utf8(p2) +  p3; }) // protect comments

                               .replace(/((?:<pre.*>))([\s\S]*?)((?:<\/pre>))/ig, function(match, p1, p2, p3) { return p1 + btoa_utf8(p2) + p3; } ) // unprotect pre content
                               .replace(/((?:<script.*>))([\s\S]*?)((?:<\/script>))/ig, function(match, p1, p2, p3) { return p1 + btoa_utf8(p2) + p3; } ) // unprotect script content

                               .replace(/\>[\t\r\n\f]*\</g, '><') // whitespaces between tags, not #32
                               .replace(/[\t\r\n\f]*\</g, '<') // whitespaces before tag, not #32
//problems: </a> a_text                               .replace(/\>[ \t\r\n\f]*/g, '>') // variant >#32text FF/Google ignores leading #32
                               .replace(/\>\s\s+/g, '>') // other white spaces after >
                               .replace(/\&nbsp\;/g, String.fromCharCode(255) + String.fromCharCode(254) + "n_bsp;") // protect &nbsp; for dom parser
                               .replace('><!--', ">\n<!--") //style sheets comment on next line
                               .replace('--></style>', "-->\n</style>") // end style sheets comment on next line
                               ;

        //                        .replace(/  +/g, ' ') // replace 2 or more spaces with one

        //console.log(htmlString);

        htmlContent = GetCleanedHTML(htmlString, options, callbackAfterDOMParser);

        htmlContent = htmlContent.replace(/\xFF\xFEn_bsp\;/g, "&nbsp;")//unprotect &nbsp;
                     .replace(/((?:<pre.*>))([\s\S]*?)((?:<\/pre>))/ig, function(match, p1, p2, p3) { return p1 + atob_utf8(p2.trim()) + p3; } ) // unprotect pre content
                     .replace(/((?:<script.*>))([\s\S]*?)((?:<\/script>))/ig, function(match, p1, p2, p3) { return p1 + atob_utf8(p2.trim()) + p3; } ) // unprotect script content
                     .replace(/((?:<\!--))([\s\S]*?)((?:-->))/g, function(match, p1, p2, p3) { return p1 + atob_utf8(p2.trim()) + p3; } ) // unprotect comments
                     .replace(/<\!\[endif\]-->/ig, '<![endif]-->' + "\r\n")
                     .replace(/<\!--\[/ig, "\r\n" + '<!--[')
                     ;

        //console.log(htmlContent);

        if(relativeUserFilesPath !== ""){
           htmlContent = htmlContent.replace(/\=\"userfiles\//g, '="' + relativeUserFilesPath);
           htmlContent = htmlContent.replace(/\&quot\;userfiles\//g, '&quot;' + relativeUserFilesPath);
        }

        return htmlContent;
}


function GetCleanedHTML( htmlString, options = { useRules: false, indentationChars: '    ', Rules: [] }, callbackAfterDOMParser ) {

  if(!options)
    options = {useRules: false, xhtml: false};

  options.useRules = options.useRules | false;

  options.xhtml = options.xhtml | false;

  if(options.useRules)
    options.Rules = options.Rules || [];

  if(options.useRules && options.Rules["pre"])
    delete options.Rules["pre"];


  let containsBody = htmlString.indexOf("<body") > -1;

  // https://www.w3schools.com/jsref/prop_node_nodetype.asp
  const ntElement = 1;
  const ntText = 3;
  const ntComment = 8;

  var myDocument = _DOMParser(htmlString);

  if(typeof callbackAfterDOMParser !== 'undefined' && callbackAfterDOMParser !== null)
    callbackAfterDOMParser(myDocument);

  var htmlContent = "";

  htmlContent += myDocument.doctype ? '<!DOCTYPE ' + myDocument.doctype.nodeName + ">\n" : " ";

  WriteHTML(myDocument.documentElement);


  if(!containsBody){
     var m = htmlContent.match(/<body[^>]*>([^<]*(?:(?!<\/?body)<[^<]*)*)<\/body\s*>/i);
     if (m) htmlContent = m[1].trim();
  }

  return htmlContent[0] == " " ? htmlContent.substr(1) : htmlContent;


  function _DOMParser( htmlString ){
     if ( typeof DOMParser !== 'undefined' && window.DOMParser.prototype.parseFromString ) {
        const parser = new DOMParser();
        return parser.parseFromString(htmlString, "text/html");
     }else{
        const newDoc = document.implementation.createHTMLDocument( 'processing doc' );
        newDoc.body.innerHTML = htmlString;
        return newDoc;
     }
  }

  function htmlElement( element ) {

     var elems = ['area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'];

     const PreserveWhiteSpaceTags = ["pre", "style"];
     const NoneLineBreakBeforeAfterTags  = "span";

     const AttribSortOrder = [];
     AttribSortOrder['table'] = "id class style height bordercolor cellspacing bordercolordark cellpadding width align bgcolor bordercolorlight background border".split(" ");
     AttribSortOrder['td'] = "id class style bgcolor height valign background width nowrap align".split(" ");
     AttribSortOrder['p'] = "id class style align name".split(" ");
     AttribSortOrder['div'] = AttribSortOrder['p'];
     AttribSortOrder['font'] = "size face".split(" ");

     this.isSingleClosingTag = function (element) {
         return elems.indexOf( element.tagName.toLowerCase() ) !== -1;
     }
     this.htmlEncodeAttr = function (text) {
         return String( text ).replace( /\&amp\;/g, '&' ).replace( /\&/g, '&amp;' ).replace( /\>/g, '&gt;' ).replace( /\</g, '&lt;' ).replace( /\"/g, '&quot;' );
     }
     this.SortAttributes = function(){
      if(AttribSortOrder[this.name]){
        let newattributes = {};
        for(var i=0; i<AttribSortOrder[this.name].length; i++){ // sort like CKEditor/SuperMailer
          if( this.attributes[ AttribSortOrder[this.name][i] ] !== undefined ){
            if( this.attributes[ AttribSortOrder[this.name][i] ] !== "" )
              newattributes[ AttribSortOrder[this.name][i] ] = this.attributes[ AttribSortOrder[this.name][i] ];
            delete this.attributes[ AttribSortOrder[this.name][i] ];
          }
        }
        for (var attr in this.attributes )
          newattributes[attr] = this.attributes[attr];
        this.attributes = newattributes;
      }
     }

     this.nodeType = element.nodeType;
     this.parentElement = element.parentElement ? element.parentElement.tagName.toLowerCase() : "";
     if(this.nodeType == ntElement){
       this.name = element.tagName.toLowerCase();
       this.attributes = {};
       for(var i=0; i<element.attributes.length; i++)
          this.attributes[element.attributes[i].nodeName] = element.attributes[i].nodeValue;
       this.isSingleClosing = this.isSingleClosingTag(element);
       this.SortAttributes();
     }else{
       this.value = element.nodeValue;
       this.isSingleClosing = true;
     }
     this.$ = element;

     this.GetOpenTagAsHTML = function(){
       if(this.nodeType == ntElement){
         var s = "";
         s += '<' + this.name;
         for (var attr in this.attributes )
           s += " " + attr + '=' + '"' + this.htmlEncodeAttr(this.attributes[attr]) + '"';
         if(!this.isSingleClosing)
           s += ">";
           else
           if( options.xhtml )
             s += " />";
             else
             s += ">";
         return s;
       }else{
         if(this.nodeType == ntComment){
           var s = "<!-- ";
           if(this.value.length > 0 && this.value[0] !== "\n")
             s += "\n";
           s += this.value;
           if(this.value.length > 0 && this.value[this.value.length - 1] !== "\n")
             s += "\n";
           s += " -->";
           return s;
           }
           else
            if( ["html", "head"].indexOf(this.parentElement) > -1 )
              return "";
            else
              return this.value;
       }
     }

     this.GetCloseTagAsHTML = function(){
       if(this.isSingleClosing || (this.nodeType !== ntElement)) return "";
       var s = "";

       s += '</' + this.name + ">";

       return s;
     }
  };



  function WriteHTML(element){

    var e = new htmlElement(element);

    CorrectHTMLElementForEMails(e);
    e.SortAttributes();

    let r = e.GetOpenTagAsHTML();

    var s = "";
    if(options.useRules && options.Rules[e.name]){
      if(options.Rules[e.name].breakBeforeOpen && htmlContent[htmlContent.length - 1] !== "\n")
         s += "\n";
      if(options.Rules[e.name].indent)
         s += options.indentationChars;
    }

    htmlContent += s + r;

    s = "";
    if(options.useRules && options.Rules[e.name])
      if(options.Rules[e.name].breakAfterOpen && htmlContent[htmlContent.length - 1] !== "\n")
         s += "\n";
    htmlContent += s;

    for(var i = 0; element.childNodes && i < element.childNodes.length; i++){
      var childelement = element.childNodes[i];

      WriteHTML(childelement);

    }

    r = e.GetCloseTagAsHTML();

    s = "";
    if(options.useRules && options.Rules[e.name])
        if(options.Rules[e.name].breakBeforeClose && htmlContent[htmlContent.length - 1] !== "\n")
           s += "\n";

    htmlContent += s + r;

    s = "";
    if(options.useRules && options.Rules[e.name])
        if(options.Rules[e.name].breakAfterClose && htmlContent[htmlContent.length - 1] !== "\n")
           s += "\n";
    htmlContent += s;

    e = null;
  }

  function CorrectHTMLElementForEMails(element){

      if(element.nodeType !== ntElement) return element;
      if(DontCorrectHTMLElementForEMails) return element;


       /**
       * Convert a CSS rgb(R, G, B) color back to #RRGGBB format.
       * @param Css style string (can include more than one color
       * @return Converted css style.
       */
      function convertRGBToHex( cssStyle )
      {
       if(cssStyle.indexOf("rgba") < 0) // chrome rgba
         return cssStyle.replace( /(?:rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\))/gi, function( match, red, green, blue )
          {
           red = parseInt( red, 10 ).toString( 16 );
           green = parseInt( green, 10 ).toString( 16 );
           blue = parseInt( blue, 10 ).toString( 16 );
           var color = [red, green, blue] ;

           // Add padding zeros if the hex value is less than 0x10.
           for ( var i = 0 ; i < color.length ; i++ )
            color[i] = String( '0' + color[i] ).slice( -2 ) ;

           return '#' + color.join( '' ) ;
           });


         return cssStyle.replace( /(?:rgba\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\))/gi, function( match, red, green, blue, alpha )
          {
           red = parseInt( red, 10 ).toString( 16 );
           green = parseInt( green, 10 ).toString( 16 );
           blue = parseInt( blue, 10 ).toString( 16 );
           var color = [red, green, blue] ;

           // Add padding zeros if the hex value is less than 0x10.
           for ( var i = 0 ; i < color.length ; i++ )
            color[i] = String( '0' + color[i] ).slice( -2 ) ;

           if(alpha == 0 && red == 0 && green == 0 && blue == 0)
             return 'transparent';
           else
             return '#' + color.join( '' ) ;
           });

      }

      if ( element.name == 'img' || element.name == 'hr' || element.name == 'table'  ){
        var style = element.attributes.style;
        if ( style )
        {
          // check for center, tinymce sets style for align=center
          const props = style.split(';');
          var alignCenter = props.length > 0;
          let alignCenterProps = ['display:block', 'margin-left:auto', 'margin-right:auto'];
          for(var i=0; i<props.length; i++){
            props[i] = props[i].trim().toLowerCase().replace(/ /ig, '');
          }
          for(var i= (element.name == 'table' ? 1 : 0); i<alignCenterProps.length; i++){
            alignCenter &= props.contains(alignCenterProps[i])
          }

          if(alignCenter)
            element.attributes.align = "center";
        }
     }

      // Output dimensions of images as width and height
      if ( element.name == 'img'  )
      {

       if(element.attributes.class != undefined && element.attributes.class === "")
         delete element.attributes.class;

       var style = element.attributes.style;

       if ( style )
       {
        // Get the width from the style.
        var match = /(?:^|\s)width\s*:\s*(\d+)px/i.exec( style );
        var width = match && match[1];

        // Get the height from the style.
        var match = /(?:^|\s)height\s*:\s*(\d+)px/i.exec( style );
        var height = match && match[1];

        // Get the float from the style.
        var match = /(?:^|\s)float\s*:\s*(\w*)(;|$)/i.exec( style );
        var afloat = match && match[1];
        if(afloat)
          afloat = afloat.replace(/;/, "");
        if(afloat != "right" && afloat != "left"){
          afload = false;
        }

        if ( width && width != "auto" )
        {
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)width\s*:\s*(\d+)px;?/i , '' );
         element.attributes.width = width;
        }

        if ( height && height != "auto" )
        {
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)height\s*:\s*(\d+)px;?/i , '' );
         element.attributes.height = height;
        }

        if ( afloat )
        {
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)float\s*:\s*(\w*);?/i , '' );
         element.attributes.align = afloat;
        }

       }
      }

      // Output alignment of paragraphs, div and td using align / valign
      if ( element.name == 'p' || element.name == 'div' || element.name == 'td' )
      {
       // delete blank background=
       if(element.attributes.background != undefined && element.attributes.background === "") {
         delete element.attributes.background;
       }

       var style = element.attributes.style;

       if ( style )
       {
        // Get the align from the style.
        var match = /(?:^|\s)text-align\s*:\s*(\w*)(;|$)/i.exec( style );
        var align = match && match[1];
        if(align)
          align = align.replace(/;/, "");

        // Get the valign from the style for td only.
        var match = /(?:^|\s)vertical-align\s*:\s*(\w*)(;|$)/i.exec( style );
        var valign = match && match[1];
        if(valign)
          valign = valign.replace(/;/, "");

        if ( align )
        {
         //element.attributes.style = element.attributes.style.replace( /(?:^|\s)text-align\s*:\s*(\w*)(;|$)?/i , '' );
         element.attributes.align = align;
        }

        if ( valign ) // for td only
        {
         //element.attributes.style = element.attributes.style.replace( /(?:^|\s)vertical-align\s*:\s*(\w*)(;|$)?/i , '' );
         element.attributes.valign = valign;
        }

       }
      }

      // Output image borders as HTML
      if ( element.name == 'img' )
      {

       if(!element.attributes.border)
         element.attributes.border = "0";

       var style = element.attributes.style;

       if ( style )
       {
        //margin *******************
        var match = /(?:^|\s)margin\s*:\s*(.*)(;|$)/i.exec( style );
        var margin = match && match[1];
        if(margin){
          if(margin.indexOf(';') > -1)
            margin = margin.substr(0, margin.indexOf(';'));
          margin = margin.replace(/;/, "");
        }
        if(margin && margin != "auto") {
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)margin\s*:\s*(.*)(;|$)/i , '' );
          margin = margin.trim();
          if(margin.indexOf(" ") != -1){ // e.g. margin 10px 20px;
            var vspace = parseInt(margin);
            var hspace = parseInt(margin.substr(margin.indexOf(" ")+1));
            element.attributes.hspace = hspace;
            element.attributes.vspace = vspace;
          } else if(parseInt(margin) > 0) { // e.g. margin 10px;
            element.attributes.hspace = parseInt(margin);
            element.attributes.vspace = parseInt(margin);
          }
        }

        // chrome long style
        var match = /(?:^|\s)margin-left\s*:\s*(\d+)px/i.exec( style );
        var margin_left = match && match[1] && parseInt(match[1]) >= 0;
        if(margin_left)
          margin_left = parseInt(match[1]);
        var match = /(?:^|\s)margin-right\s*:\s*(\d+)px/i.exec( style );
        var margin_right = match && match[1] && parseInt(match[1]) >= 0;
        if(margin_right)
          margin_right = parseInt(match[1]);
        var match = /(?:^|\s)margin-top\s*:\s*(\d+)px/i.exec( style );
        var margin_top = match && match[1] && parseInt(match[1]) >= 0;
        if(margin_top)
          margin_top = parseInt(match[1]);
        var match = /(?:^|\s)margin-bottom\s*:\s*(\d+)px/i.exec( style );
        var margin_bottom = match && match[1] && parseInt(match[1]) >= 0;
        if(margin_bottom)
          margin_bottom = parseInt(match[1]);

        if(margin_left && margin_right && (margin_left == margin_right)){
          element.attributes.hspace = margin_left;
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)margin-left\s*:\s*(\d+)px(;|$)/i , '' );
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)margin-right\s*:\s*(\d+)px(;|$)/i , '' );
        }

        if(margin_top && margin_bottom && (margin_top == margin_bottom)){
          element.attributes.vspace = margin_top;
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)margin-top\s*:\s*(\d+)px(;|$)/i , '' );
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)margin-bottom\s*:\s*(\d+)px(;|$)/i , '' );
        }

        // border *************************

        // FF short style
        // border-width.
        var match = /(?:^|\s)border-width\s*:\s*(\d+)px/i.exec( style );
        var border_width = match && match[1] && parseInt(match[1]) >= 0;
        if(border_width)
          border_width = parseInt(match[1]);
        // border-style
        var match = /(?:^|\s)border-style\s*:\s*(\w*)(;|$)/i.exec( style );
        var border_style = match && match[1];
        if(border_style)
          border_style = border_style.replace(/;/, "");
          else
          if(border_width && border_width > 0)
             border_style = "solid";

        // IE long style
        var match = /(?:^|\s)border-bottom\s*:\s*(\S*)\s*solid/i.exec( style );
        var border_bottom = match && match[1] && match[0].indexOf("solid") != -1 && parseInt(match[1]) >= 0;
        if(border_bottom)
          border_bottom = parseInt(match[1]);

        var match = /(?:^|\s)border-left\s*:\s*(\S*)\s*solid/i.exec( style );
        var border_left = match && match[1] && match[0].indexOf("solid") != -1 && parseInt(match[1]) >= 0;
        if(border_left)
          border_left = parseInt(match[1]);

        var match = /(?:^|\s)border-right\s*:\s*(\S*)\s*solid/i.exec( style );
        var border_right = match && match[1] && match[0].indexOf("solid") != -1 && parseInt(match[1]) >= 0;
        if(border_right)
          border_right = parseInt(match[1]);

        var match = /(?:^|\s)border-top\s*:\s*(\S*)\s*solid/i.exec( style );
        var border_top = match && match[1] && match[0].indexOf("solid") != -1 && parseInt(match[1]) >= 0;
        if(border_top)
          border_top = parseInt(match[1]);

        if(border_bottom && border_left && border_right && border_top){

          var border = new Array(border_bottom, border_left, border_right, border_top);
          var b = border[0];
          for(var i=1; i<border.length; i++)
            if(b != border[i]){
              b = -1;
              break;
            }

          if(b != -1) {
            border_width = String(parseInt(border_bottom));
            border_style = "solid";
          }
        }
        // IE long style /

        // Chrome long style
        var match = /(?:^|\s)border-top-width\s*:\s*(\d+)px/i.exec( style );
        var border_top = match && match[1] && parseInt(match[1]) >= 0;
        if(border_top)
          border_top = parseInt(match[1]);

        var match = /(?:^|\s)border-right-width\s*:\s*(\d+)px/i.exec( style );
        var border_right = match && match[1] && parseInt(match[1]) >= 0;
        if(border_right)
          border_right = parseInt(match[1]);

        var match = /(?:^|\s)border-left-width\s*:\s*(\d+)px/i.exec( style );
        var border_left = match && match[1] && parseInt(match[1]) >= 0;
        if(border_left)
          border_left = parseInt(match[1]);

        var match = /(?:^|\s)border-bottom-width\s*:\s*(\d+)px/i.exec( style );
        var border_bottom = match && match[1] && parseInt(match[1]) >= 0;
        if(border_bottom)
          border_bottom = parseInt(match[1]);

        var match = /(?:^|\s)border-top-style\s*:\s*(\w*)(;|$)/i.exec( style );
        var border_style_top = match && match[1] && match[1].lastIndexOf('solid') != -1;
        var match = /(?:^|\s)border-right-style\s*:\s*(\w*)(;|$)/i.exec( style );
        var border_style_right = match && match[1] && match[1].lastIndexOf('solid') != -1;
        var match = /(?:^|\s)border-left-style\s*:\s*(\w*)(;|$)/i.exec( style );
        var border_style_left = match && match[1] && match[1].lastIndexOf('solid') != -1;
        var match = /(?:^|\s)border-bottom-style\s*:\s*(\w*)(;|$)/i.exec( style );
        var border_style_bottom = match && match[1] && match[1].lastIndexOf('solid') != -1;

        if(border_style_top && border_style_right && border_style_left && border_style_bottom)
          border_style = "solid";


        if(border_bottom && border_left && border_right && border_top){

          var border = new Array(border_bottom, border_left, border_right, border_top);
          var b = border[0];
          for(var i=1; i<border.length; i++)
            if(b != border[i]){
              b = -1;
              break;
            }

          if(b != -1) {
            border_width = String(parseInt(border_bottom));
          }
        }
        // Chrome long style /


        if ( border_width >= 0 && border_style == "solid" )
        {

         // FF
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-width\s*:\s*(\d+)px?/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-style\s*:\s*(\w*)(;|$)?/i , '' );

         // IE long style with/without ;
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-bottom\s*:\s*(\S*)\s*solid(;|$)/i, '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-left\s*:\s*(\S*)\s*solid(;|$)/i, '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-right\s*:\s*(\S*)\s*solid(;|$)/i, '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-top\s*:\s*(\S*)\s*solid(;|$)/i, '' );
         // // IE long style with/without ; /

         // Chrome long style
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-top-width\s*:\s*(\d+)px(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-right-width\s*:\s*(\d+)px(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-left-width\s*:\s*(\d+)px(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-bottom-width\s*:\s*(\d+)px(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-top-style\s*:\s*(\w*)(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-right-style\s*:\s*(\w*)(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-left-style\s*:\s*(\w*)(;|$)/i , '' );
         element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-bottom-style\s*:\s*(\w*)(;|$)/i , '' );
         // Chrome long style//

         element.attributes.border = parseInt(border_width);
        }
       }
      }

      // output td background-color as bgcolor and white-spacing als nowrap
      if ( element.name == 'table' || element.name == 'td' )
      {
       // IE bug <td align="middle"> is invalid must be <td align="center">
       if(element.attributes.align && element.attributes.align == "middle")
         element.attributes.align = "center";

       // delete blank background=
       if(element.attributes.background != undefined && element.attributes.background === "") {
         delete element.attributes.background;
       }

       var style = element.attributes.style;
       if ( style )
       {

         style = convertRGBToHex( style ); // FF rgb() problem
         element.attributes.style = style;

         // Get the width from the style.
         var match = /(?:^|\s)width\s*:\s*(\S*)?/i.exec( style );
         var width = match && match[1];
         if(width && width.indexOf("%") == -1 && width != "auto")
           width = parseInt(width);

         // Get the height from the style.
         var match = /(?:^|\s)height\s*:\s*(\S*)?/i.exec( style );
         var height = match && match[1];
         if(height && height.indexOf("%") == -1 && height != "auto")
           height = parseInt(height);

         if ( width && width != "auto" )
         {
          width = String(width).replace(/;/, "");
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)width\s*:\s*(\S*)?\;/i , '' ).replace( /(?:^|\s)width\s*:\s*(\S*)?/i , '' );

          if( width.indexOf("%") > -1 ){
             width = Math.round( parseFloat(width.substr(0, width.indexOf("%"))) ) + "%";
          }

          element.attributes.width = width;
         }

         if ( height && height != "auto" )
         {
          height = String(height).replace(/;/, "");
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)height\s*:\s*(\S*)?\;/i , '' ).replace( /(?:^|\s)height\s*:\s*(\S*)?/i , '' );
          if(element.name !== 'table')
            element.attributes.height = height;
            else
            if(element.attributes.height)
               delete element.attributes.height; // in tables no height
         }

         // background-color
         var match = /(?:^|\s)background-color\s*:\s*(\S*)(;|$)/i.exec( style );
         var background_color = match && match[1];

         if(background_color){
           // we let it in td tag, so webbased emailers can't override it --> element.attributes.style = element.attributes.style.replace( /(?:^|\s)background-color\s*:\s*(\S*)(;|$)?/i , '' );
           background_color = background_color.replace(/;/, "");
           element.attributes.bgcolor = background_color;
         }

         // white-space: nowrap
         var match = /(?:^|\s)white-space\s*:\s*(\S*)(;|$)/i.exec( style );
         var white_space = match && match[1];
         if(white_space)
           white_space = white_space.replace(/;/, "");
         if(white_space != "nowrap")
            white_space = false;
         if(white_space){
           element.attributes.style = element.attributes.style.replace( /(?:^|\s)white-space\s*:\s*(\S*)(;|$)?/i , '' );
           element.attributes.nowrap = "nowrap";
         }

         if(element.name == 'table'){

           var style = element.attributes.style;

           var match = /(?:^|\s)border-width\s*:\s*(\d+)px/i.exec( style );
           var border_width = match && match[1] && parseInt(match[1]) >= 0;
           if(border_width){
             border_width = parseInt(match[1]);
             element.attributes.border = border_width;
             element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-width\s*:\s*(\d+)px?/i , '' );
           }

           var match = /(?:^|\s)border-spacing\s*:\s*(\d+)px/i.exec( style );
           var border_spacing = match && match[1] && parseInt(match[1]) >= 0;
           if(border_spacing){
             border_spacing = parseInt(match[1]);
             element.attributes.cellspacing = border_spacing;
             element.attributes.style = element.attributes.style.replace( /(?:^|\s)border-spacing\s*:\s*(\d+)px?/i , '' );
           }

           if(element.$ && element.$.childNodes.length > 0){
             var childs = element.$.childNodes;
             var tbody, tr, td;
             for(var i=0; i<childs.length; i++){
              if(childs[i].nodeType != 1) continue;
              if(childs[i].tagName == "TBODY"){
                tbody = childs[i];
                break;
              }
              if(childs[i].tagName == "TR"){
                tr = childs[i];
                break;
              }
             }
             if(tbody){
               var childs = tbody.childNodes;

               for(var i=0; i<childs.length; i++){
                if(childs[i].nodeType != 1) continue;

                if(childs[i].tagName == "TR"){
                  tr = childs[i];
                  td = tr.cells[0];
                  break;
                }
               }
             }else if(tr){
               td = tr.cells[0];
             }

             if(td && td.attributes && td.attributes.style){
               var style = td.attributes.style.nodeValue;
               var match = /(?:^|\s)padding\s*:\s*(\d+)px/i.exec( style );
               var padding = match && match[1] && parseInt(match[1]) >= 0;
               if(padding){
                 padding = parseInt(match[1]);
                 element.attributes.cellpadding = padding;
               }
             }
           }

         }

       }
      }

      if ( element.name == 'tr' || element.name == 'td'  ){
        var style = element.attributes.style;

        // Get the height from the style.
        var match = /(?:^|\s)height\s*:\s*(\S*)?/i.exec( style );
        var height = match && match[1];

        if(height && height !== "auto")
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)height\s*:\s*(\S*)?\;/i , '' ).replace( /(?:^|\s)height\s*:\s*(\S*)?/i , '' );
        if(element.attributes.height)
           delete element.attributes.height; // in tr and td no height


         // Get the width from the style.
         var match = /(?:^|\s)width\s*:\s*(\S*)?/i.exec( style );
         var width = match && match[1];
         if(width && width.indexOf("%") == -1)
           width = parseInt(width);

         if ( width && width !== "auto" )
         {
          width = String(width).replace(/;/, "");
          element.attributes.style = element.attributes.style.replace( /(?:^|\s)width\s*:\s*(\S*)?\;/i , '' ).replace( /(?:^|\s)width\s*:\s*(\S*)?/i , '' );
          if( element.name == 'tr' )
            delete element.attributes.width;

          element.attributes.width = width;
         }


      }

      // output body background-color as bgcolor
      if ( element.name == 'body' && element.attributes.style )
      {
       var style = element.attributes.style;
       if ( style )
       {
         style = convertRGBToHex( style ); // FF rgb() problem
         element.attributes.style = style;
         // background-color
         var match = /(?:^|\s)background-color\s*:\s*(\S*)(;|$)/i.exec( style );
         var background_color = match && match[1];

         if(background_color){
           // we let it in body tag, so webbased emailers can't override it --> element.attributes.style = element.attributes.style.replace( /(?:^|\s)background-color\s*:\s*(\S*)(;|$)?/i , '' );
           background_color = background_color.replace(/;/, "");
           element.attributes.bgcolor = background_color;
         }

       }
      }

      if ( !element.attributes.style || element.attributes.style.replace(/ /, "") == "" || element.attributes.style.replace(/ /, "") == ";" )
       delete element.attributes.style;
       else {
         // FF problems with &quot; in url()
         var style = element.attributes.style;
         if(style.indexOf("url(&quot;") != -1) {
            style = style.replace("url(&quot;", "url('");
            style = style.replace("&quot;)", "')");
            element.attributes.style = style;
         }
       }


      if(element.attributes && element.attributes.style){
        var style = element.attributes.style;
        style = convertRGBToHex( style );
        element.attributes.style = style;

        element.attributes.style = element.attributes.style.replace( /\;\;/g, '' ).replace( /"\;\s*\;"/g, '' ).trimStart();
        if(element.attributes.style === "")
          delete element.attributes.style;
      }

      return element;
  }

}

function CheckSecurityOfPastedHTMLCode(htmlcode, editor){

  let invalidTags = [
   'o:p',
   'xml',
   'script',
   'meta',
   'link',
   'iframe',
   'media',
   'embed',
   'object'
  ];



  htmlcode = htmlcode.replace(/class\=\"MsoNormal\"/ig, '');
  htmlcode = htmlcode.replace(/class\=\"MsoBodyText\"/ig, '');

  htmlcode = htmlcode.replace(/^\?xml:namespace$/ig, '');
  htmlcode = htmlcode.replace(/^v:shapetype/ig, '');

  for(var i=0; i<invalidTags.length; i++){
    if(editor.config.allow_script_urls && invalidTags[i] == "script") continue;
    htmlcode = htmlcode.replaceAll( new RegExp("<" + invalidTags[i] + ".*>.*?<\/" + invalidTags[i] + ">", "gi"), '');
  }

  htmlcode = htmlcode.replace(/\\\* MERGEFORMAT/ig, '');

  // Sometimes Word malforms the comments.
  htmlcode = htmlcode.replaceAll( /<!\[/g, '<!--[' ).replace( /\]>/g, ']-->' );

  return htmlcode;
}

